//****************************************************************
//*                                                              *
//*    Expr.h : Header file of the Expression hierarchy classes  *
//*             This file describes all the nodes contained into *
//*             an expression tree of the program.               *
//*                                                              *
//* (c) LITA, university of PAU (France), summer 1993.           *
//****************************************************************

//**************************************
//**                                  **
//**  Objects decribing an expression **
//**  inside an abstract tree.        **
//**                                  **
//**************************************

// This is a global variable used for identify the location of a lexem.
extern Location ThisPlace;

//**
//** Enumeration type used for calculating the return type
//** of one expression node.
//**

typedef enum
{
  Ident,        // Identifier node : could match any kind of expression
  IntegerConst, // Integer constant node.
  IntegerExpr,  // Integer expression node.
  RealConst,    // Real constant node.
  RealExpr,     // Real expression node.
  BoolConst,    // Boolean constant node.
  BoolExpr,     // Boolean expression node.
  ObjConst,     // Object constant node (like NONE operator or
                // predefined type ).
  ObjExpr,      // Object expression node.
  StringExpr,   // String expression node.
  StringConst,  // String constant node.
  CharExpr,     // Character expression node.
  CharConst,    // Character constant node.
  Other,        // Other Node (used for constructors, will be removed too ).
  ExprError     // This node and maybe its children is holding
                // a wrong expression.
} ExprType;

// We overload this operator for let it print an ExprType object.
ostream& operator << ( ostream& , ExprType    );

//**
//** This class is a dummy class that is just for testing code.
//** The code should be removed when joining to the symbol Table.
//** This is supposed to described the type of one symbol table entry.
//** The only node to use it is the Identifier Node for now.
//**

class Entry
{
  int dummy; //** Just to say the class is not empty.
};

//**
//** This is the Expression class. This class is the root of all the tree
//** of the expression classes. You find in it all the method and fields
//** that every node uses.
//**

class Expression
{

  public:
  String       kind;       // Verbose description of the type of the node.
  Location    *place;      // Position of the object in the source text.
  ExprType     returntype; // Kind of result the node produces.

  Expression( ExprType,Location* );// The constructor take just the kind
                                   // of the node, and eventually its position.
  virtual void Print( ostream& );  // Virtual procedure that prints the contents
                                   // of the node. Of course there is a lot of
                                   // different version of this procedure.

// This overloading allow the programmer to write 'cout << TheExpression'
// for printing the contents of all the nodes in the tree.
  friend ostream& operator << ( ostream& , Expression );
} ;

//**
//** This is the Identifier Operator Class that derives from Expression
//** class. This nodes holds a string field to store temporaly the string that
//** lex has parsed and an entry type for the variable in the symbol table.
//**

class Identifier : public Expression
{
  public:
  String *Title; // The name of the operator.
  Entry  Where;  // The entry if defined (else NULL).

  Identifier( String *, Location * );
                          // Initialised with the name if it has not been
                          // declared (maybe further).
  Identifier( Entry, Location *);    // else initialised with its entry.
  void Print( ostream& ); // This Print write down the name of the identifier.

} ;

//**
//** This is the UnaryOperator class. This class derives from Expression
//** from which it adds an Expression pointer for one argument of type
//** Expression. This class in general should not be instantiated.
//** Use derivation for creating new unary operator.
//**

class UnaryOperator : public Expression
{
  public:
  Expression *Argument;  // This holds the argument which is another
                         // expression.

// We need to construct this the argument taken and what kind of result
// this operator returns.

  UnaryOperator( Expression *ThisExp, ExprType ret, Location * ); 

  virtual void Print( ostream& ); // This print outputs the name of
                                  // the operator and the argument into brackets.

} ;

//**
//** This is the BinaryOperator class. This is the same as UnaryOperator
//** except it takes two expression instead of one.
//** This class is separate from the UnaryOperator because functions that
//** takes one argument can't accept two arguments (which will surely happend
//** if BinaryOperator derives from UnaryOperator ).
//**

class BinaryOperator : public Expression
{
public:
  Expression *LeftMember,*RightMember;

  BinaryOperator( Expression *LeftExpr, Expression *RightExpr,
                  ExprType    ret,      Location   *TheLoc );
  virtual void Print( ostream& ); // The two expression are printed into
                                  // brackets and separated by a ','.

} ;

//**
//** This is the BoolOperator class. This class describes all the boolean
//** operators ( =, =/= or <>, <, <=, >, >=, OR, AND, ORIF and ANDIF ).
//** The operators are discimined by the field OpKind.
//** We use for this a new enumeration type called BoolOpType (BOOLean OPerator
//** TYPE).
//**

typedef enum
{
  Less,            // A <     B
  LessOrEqual,     // A <=    B
  Equal,           // A =     B
  NotEqual,        // A =/=   B or A <> B
  Greater,         // A >     B
  GreaterOrEqual,  // A >=    B
  Or,              // A OR    B (inclusif OR)
  And,             // A AND   B
  AndIf,           // A ANDIF B (same as AND)
  OrIf             // A ORIF  B (same as OR)
} BoolOpType;

// We overload << for printing BoolOpType variable.

ostream& operator << ( ostream& , BoolOpType  );

class BoolOperator : public BinaryOperator
{
public:
  BoolOpType OpKind;

// For creation, we need the two expression and the kind of boolean operator.
// For example if a is an 'BoolOperator *'
// then 'a = new BoolOperator( Expr1, Expr2, Or )' will describe the expression
// '(Expr1) OR (Expr2)'.
  BoolOperator( Expression *LeftExpr, Expression *RightExpr,
                BoolOpType  Operator, Location * );
  void Print( ostream& );
} ;

//**
//** This is the ArithOperator class. This class is designed on the same model
//** as BoolOperator but for arithmetic Operator.
//** The new enumerated type is calls ArithOpType (ARITHmetic OPerator TYPE).
//**

typedef enum
{
  Plus,         // A  +  B
  Minus,        // A  -  B
  Multiply,     // A  *  B
  Divide,       // A  /  B
  IntDivide,    // A DIV B
  Modulo        // A MOD B
} ArithOpType;

ostream& operator << ( ostream& , ArithOpType );

class ArithOperator : public BinaryOperator
{
public:
  ArithOpType OpKind;
  ArithOperator( Expression  *LeftExpr, Expression  *RightExpr,
                 ArithOpType  Operator, Location * );
  void Print( ostream& );
} ;

//**
//** This is the ObjOperator class. This class derives from the class
//** BinaryOperator. It describes the operators which need object expression
//** as argument (essentially QUA, IS ,IN ).
//**

typedef enum
{
  Qua,
  Is,
  In
} ObjOpType;

ostream& operator << ( ostream& , ObjOpType   );

class ObjOperator   : public BinaryOperator
{
public:
  ObjOpType OpKind;
  ObjOperator( Expression *LeftExpr, Expression *RightExpr,
               ObjOpType   Operator, Location * );
  void Print( ostream& );
} ;

class Not : public UnaryOperator
{
public:
  Not( Expression *TheExpr, Location * );
} ;

class This : public UnaryOperator
{
  public:
  This( Expression *TheExpr, Location * );
  void Print( ostream& );
} ;

class New : public UnaryOperator
{
  public:
  New( Expression *TheExpr, Location * );
} ;

class Copy : public UnaryOperator
{
  public:
  Copy( Expression *TheExpr, Location * );
} ;

class ArrayOf : public UnaryOperator
{
  public:
  ArrayOf( Expression *TheExpr, ExprType ret, Location * );
} ;

class NoneObject : public Expression
{
  public:
  NoneObject( Location * );
} ;

class Result : public Expression
{
  public:
  Result( Location * );
} ;

class Error : public Expression
{
  public:
  Error( void );
} ;

class IntegerConstant : public Expression
{
public:
  int rvalue;
  IntegerConstant( int value, Location * );
  void Print( ostream& );

} ;

class RealConstant : public Expression
{
public:
  double rvalue;
  RealConstant( double value, Location * );
  void Print( ostream& );

} ;

class BoolConstant : public Expression
{
  public:
  int rvalue;

  BoolConstant( int, Location * );
  void Print( ostream& );
} ;

class StringConstant : public Expression
{
  public:
  String *rvalue;

  StringConstant( String *, Location * );
  void Print( ostream& );
} ;

class CharConstant : public Expression
{
  public:
  char rvalue;

  CharConstant( char rvalue, Location * );
  void Print( ostream& );
} ;
